from openfisca_us.model_api import *


class ny_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY CTC"
    documentation = "New York's Empire State Child Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c-1)
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ny.tax.income.credits.ctc
        # Qualifying children.
        person = tax_unit.members
        qualifies_for_federal_ctc = person("is_ctc_qualifying_child", period)
        age = person("age", period)
        qualifies = qualifies_for_federal_ctc & (age >= p.minimum_age)
        qualifying_children = tax_unit.sum(qualifies)
        # First calculate federal match.
        federal_match = tax_unit("ny_federal_ctc", period) * p.amount.percent
        # Scale federal match by number of qualifying children.
        federal_qualifying_children = tax_unit.sum(qualifies_for_federal_ctc)
        qualifying_federal_match = where(
            federal_qualifying_children > 0,
            federal_match * qualifying_children / federal_qualifying_children,
            0,
        )
        # Filers with income below the CTC phase-out threshold receive a
        # minimum amount per child.
        minimum = p.amount.minimum * qualifying_children
        agi = tax_unit("adjusted_gross_income", period)
        ctc = parameters(
            "2017-01-01"
        ).gov.irs.credits.ctc  # Uses pre-TCJA parameters
        filing_status = tax_unit("filing_status", period)
        federal_threshold = ctc.phase_out.threshold[filing_status]
        eligible_for_minimum = agi < federal_threshold
        applicable_minimum = eligible_for_minimum * minimum
        eligible = qualifying_children > 0
        return eligible * max_(applicable_minimum, qualifying_federal_match)



from openfisca_us.model_api import *


class lives_with_joint_filing_spouse(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Indicates whether the tax filing head lived with their joint-filing spouse the entire year. False if not married filing jointly."

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        return filing_status == filing_status.possible_values.JOINT



from openfisca_us.model_api import *


class ny_federal_actc(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY federal ACTC"
    unit = USD
    documentation = "The version of the federal ACTC used to determine the NY Empire State Child Credit."
    definition_period = YEAR
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        ctc = parameters(
            "2017-01-01"
        ).gov.irs.credits.ctc  # Federal ACTC, but with pre-TCJA parameters

        total_ctc = tax_unit("ctc", period)
        maximum_refundable_ctc = min_(
            add(tax_unit, period, ["ctc_refundable_individual_maximum"]),
            total_ctc,
        )

        # The other part of the "lesser of" statement is: "the amount by which [the non-refundable CTC]
        # would increase if [tax liability] increased by tax_increase", where tax_increase is the greater of:
        # - the phase-in amount
        # - Social Security tax minus the EITC
        # First, we find tax_increase:

        earnings = add(tax_unit, period, ["earned_income"])
        earnings_over_threshold = max_(
            0, earnings - ctc.refundable.phase_in.threshold
        )
        relevant_earnings = (
            earnings_over_threshold * ctc.refundable.phase_in.rate
        )

        # Compute "Social Security taxes" as defined in the US Code for the ACTC.
        # This includes OASDI and Medicare payroll taxes, as well as half
        # of self-employment taxes.
        PERSON_VARIABLES = [
            "employee_social_security_tax",
            "employee_medicare_tax",
            "unreported_payroll_tax",
        ]
        PERSON_VARIABLES_SUBTRACT = ["excess_payroll_tax_withheld"]
        TAX_UNIT_VARIABLES = [
            "c03260",  # Deductible portion of the self-employed tax.
            "additional_medicare_tax",
        ]
        social_security_tax = (
            add(tax_unit, period, PERSON_VARIABLES)
            + add(tax_unit, period, TAX_UNIT_VARIABLES)
            - add(tax_unit, period, PERSON_VARIABLES_SUBTRACT)
        )
        eitc = tax_unit("eitc", period)
        social_security_excess = max_(0, social_security_tax - eitc)
        qualifying_children = add(
            tax_unit, period, ["is_ctc_qualifying_child"]
        )
        tax_increase = where(
            qualifying_children
            < ctc.refundable.phase_in.min_children_for_ss_taxes_minus_eitc,
            relevant_earnings,
            max_(relevant_earnings, social_security_excess),
        )
        limiting_tax = tax_unit("ctc_limiting_tax_liability", period)
        ctc_capped_by_tax = min_(total_ctc, limiting_tax)
        ctc_capped_by_increased_tax = min_(
            total_ctc, limiting_tax + tax_increase
        )
        amount_ctc_would_increase = (
            ctc_capped_by_increased_tax - ctc_capped_by_tax
        )

        ny_federal_actc = min_(
            maximum_refundable_ctc, amount_ctc_would_increase
        )

        remaining_unused_ctc = tax_unit(
            "ny_federal_ctc_max", period
        ) - tax_unit("ny_federal_non_refundable_ctc", period)
        return min_(ny_federal_actc, remaining_unused_ctc)



from openfisca_us.model_api import *


class mo_federal_income_tax_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "MO Federal income tax deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-1040%20Instructions_2021.pdf#page=8",
        "https://revisor.mo.gov/main/OneSection.aspx?section=143.171&bid=49937&hl=federal+income+tax+deduction%u2044",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        mo_adjusted_gross_income = tax_unit("mo_adjusted_gross_income", period)
        federal_tax = tax_unit("income_tax", period)

        # subtract CARES act credits, only affects year 2020, source: https://revisor.mo.gov/main/OneSection.aspx?section=143.171&bid=48731
        cares_rebate = tax_unit("rrc_cares", period)
        federal_tax_less_cares = federal_tax - cares_rebate

        filing_status = tax_unit("filing_status", period)

        federal_income_tax_deduction_rates = parameters(
            period
        ).gov.states.mo.tax.income.deductions.federal_income_tax_deduction_rates
        rate = federal_income_tax_deduction_rates.calc(
            mo_adjusted_gross_income
        )
        federal_income_tax_deduction_amount = federal_tax_less_cares * rate

        federal_income_tax_deduction_cap = parameters(
            period
        ).gov.states.mo.tax.income.deductions.federal_income_tax_deduction_caps[
            filing_status
        ]

        return min_(
            federal_income_tax_deduction_amount,
            federal_income_tax_deduction_cap,
        )



from openfisca_us.model_api import *


class pays_property_tax_or_rent(Variable):
    value_type = float
    entity = TaxUnit
    label = "Household pays rent or property tax"
    unit = USD
    definition_period = YEAR
    reference = "https://dor.mo.gov/forms/MO-PTS_2021.pdf"

    def formula(tax_unit, period, parameters):
        return add(tax_unit, period, ["rent", "real_estate_taxes"]) > 0



from openfisca_us.model_api import *


class mo_property_tax_credit_demographic_tests(Variable):
    value_type = float
    entity = TaxUnit
    label = "MO property tax credit demographic eligiblity test"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-PTS_2021.pdf",
        "https://revisor.mo.gov/main/OneSection.aspx?section=135.010&bid=6435",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        # Eligibility
        # Check for age eligiblity
        age_head = tax_unit("age_head", period)
        age_spouse = tax_unit("age_spouse", period)
        age_threshold = parameters(
            period
        ).gov.states.mo.tax.income.credits.property_tax.age_threshold
        elderly_head = age_head >= age_threshold
        elderly_spouse = age_spouse >= age_threshold
        elderly_head_or_spouse = elderly_head | elderly_spouse

        # Check for disability eligibility
        disabled_head = tax_unit("disabled_head", period)
        disabled_spouse = tax_unit("disabled_spouse", period)
        disabled_head_or_spouse = disabled_head | disabled_spouse

        # Check for military disabled eligibility
        military_disabled_head = tax_unit("military_disabled_head", period)
        military_disabled_spouse = tax_unit("military_disabled_spouse", period)
        military_disabled_head_or_spouse = (
            military_disabled_head | military_disabled_spouse
        )

        # Check for receipt of surviving spouse benefits
        receives_survivor_benefits = (
            add(tax_unit, period, ["social_security_survivors"]) > 0
        )

        return (
            elderly_head_or_spouse
            | disabled_head_or_spouse
            | military_disabled_head_or_spouse
            | receives_survivor_benefits
        )



from openfisca_us.model_api import *


class ny_federal_ctc_max(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY federal CTC maximum"
    unit = USD
    documentation = "The version of the federal CTC used to determine the NY Empire State Child Credit, before limiting to tax liability."
    definition_period = YEAR
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        ctc = parameters(
            "2017-01-01"
        ).gov.irs.credits.ctc  # Uses pre-TCJA parameters
        count_children = add(tax_unit, period, ["is_ctc_qualifying_child"])
        filing_status = tax_unit("filing_status", period)
        maximum_amount = ctc.child.amount * count_children
        phase_out_threshold = ctc.phase_out.threshold[filing_status]
        agi = tax_unit("adjusted_gross_income", period)
        income_over_threshold = max_(0, agi - phase_out_threshold)
        reduction = income_over_threshold * ctc.phase_out.rate
        return max_(0, maximum_amount - reduction)



from openfisca_us.model_api import *


class no_salt_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Federal income tax if SALT were abolished"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        with BranchedSimulation(tax_unit) as simulation:
            simulation.tax_benefit_system.neutralize_variable("salt_deduction")
            simulation.get_holder(
                "salt_deduction"
            ).variable = simulation.tax_benefit_system.variables[
                "salt_deduction"
            ]  # This is needed because neutralize_variable is only designed to work *before* simulations have started.
            tax_liability = simulation.calculate("income_tax", period)
            simulation.get_holder(
                "salt_deduction"
            ).variable.is_neutralized = False

        return tax_liability



from openfisca_us.model_api import *


class mo_property_tax_credit_amount(Variable):
    value_type = float
    entity = TaxUnit
    label = "MO property tax credit demographic eligiblity test"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-PTS_2021.pdf",
        "https://revisor.mo.gov/main/OneSection.aspx?section=135.025&bid=6438",
        "https://revisor.mo.gov/main/OneSection.aspx?section=135.030&bid=6439",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        return 0
        # short circuit until fully modeled
        # Currently not including railroad retirement or veterans benefits.
        rents = tax_unit("rents", period)
        cohabitates = tax_unit("lives_with_joint_filing_spouse", period)
        p = parameters(period).gov.states.mo.tax.income.credits.property_tax
        income_threshold = select(
            [
                rents & cohabitates,
                rents & ~cohabitates,
                ~rents & cohabitates,
                ~rents & ~cohabitates,
            ],
            [
                p.rent_cohabitating,
                p.rent_separate,
                p.own_cohabitating,
                p.own_separate,
            ],
        )

        # Determine if the tax unit head and spouse co-habitate

        rent = add(tax_unit, period, ["rent"])
        property_tax = add(tax_unit, period, ["real_estate_taxes"])

        total_household_income = add(
            tax_unit,
            period,
            [
                "mo_adjusted_gross_income",
                "pension_income",
                "mo_property_tax_credit_public_assistance",
            ],
        )

        meets_income_test = total_household_income <= income_threshold

        rent_expense_limit = p.rental_expense_cap
        rent_total = min_(rent, rent_expense_limit)

        property_tax_expense_limit = p.property_tax_expense_cap
        property_tax_total = min_(property_tax, property_tax_expense_limit)

        # Total credit basis comes from line 13 of MO-PTS, not in legislation, proscribes $1,100 cap just as property tax expense cap
        total_credit_basis = where(
            (rent_total + property_tax_total >= property_tax_expense_limit),
            property_tax_expense_limit,
            (rent_total + property_tax_total),
        )

        minimum_base = p.minimum_base

        # Check if rent/property tax > 1100
        # placehold to complete unit testing while waiting for comment from MO dept of revenue
        return total_credit_basis



from openfisca_us.model_api import *


class is_young_child_for_medicaid_nfc(Variable):
    value_type = bool
    entity = Person
    label = "Medicaid young child non-financial criteria"
    definition_period = YEAR

    def formula(person, period, parameters):
        ma = parameters(
            period
        ).gov.hhs.medicaid.eligibility.categories.young_child
        age = person("age", period)
        return ma.age_range.calc(age)



from openfisca_us.model_api import *


class mo_adjusted_gross_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Missouri adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = (
        "",
        "",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        agi = tax_unit("adjusted_gross_income", period)
        subtractions = tax_unit(
            "mo_qualified_health_insurance_premiums", period
        )
        return max_(agi - subtractions, 0)



from openfisca_us.model_api import *


class exemptions(Variable):
    value_type = int
    entity = TaxUnit
    label = "Number of tax exemptions"
    unit = USD
    definition_period = YEAR

    formula = sum_of_variables(["tax_unit_size"])



from openfisca_us.model_api import *


class mo_property_tax_credit_public_assistance(Variable):
    value_type = float
    entity = TaxUnit
    label = "MO property tax credit demographic eligiblity test"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-PTS_2021.pdf",
        "https://dor.mo.gov/forms/4711_2021.pdf",
        "https://revisor.mo.gov/main/OneSection.aspx?section=135.010&bid=6435&hl=property+tax+credit%u2044",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        # The second reference specifies that food stamps are not counted as income for this form.
        # Form also specifies that verterans payments and benefits are included unless an individual
        # is 100% disabled as a result of military service
        # fully_disabled_service_connected_veteran = person("is_fully_disabled_service_connected_veteran", period)
        # unclear if we currently model veterans benefits
        # tanf is currently not included in this formula, issue here: https://github.com/PolicyEngine/openfisca-us/issues/1361
        return add(tax_unit, period, ["ssi", "child_support_received"])



from openfisca_us.model_api import *


class pa_total_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "PA total taxable income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2021/2021_pa-40in.pdf#page=8"
    defined_for = StateCode.PA

    formula = sum_of_variables(["adjusted_gross_income"])



from openfisca_us.model_api import *


class mo_property_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "MO property tax credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-PTS_2021.pdf",
        "https://revisor.mo.gov/main/OneSection.aspx?section=135.010&bid=6435",
        "https://revisor.mo.gov/main/OneSection.aspx?section=135.030&bid=6439",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):

        # Check demographic eligibility
        demographic_qualification = tax_unit(
            "mo_property_tax_credit_demographic_tests", period
        )
        # Check for housing expense eligibility

        any_housing_cost = tax_unit("pays_property_tax_or_rent", period)
        # Check demographic eligibility.
        credit = tax_unit("mo_property_tax_credit_amount", period)
        return demographic_qualification * any_housing_cost * credit



from openfisca_us.model_api import *


class mo_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Missouri AGI minus deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-A_2021.pdf",
        "https://www.revisor.mo.gov/main/OneSection.aspx?section=143.111&bid=7201&hl=",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        tax_unit_itemizes = tax_unit("tax_unit_itemizes", period)

        mo_agi = tax_unit("mo_adjusted_gross_income", period)
        mo_federal_income_tax_deduction = tax_unit(
            "mo_federal_income_tax_deduction", period
        )

        # MO standard deduction is set equal to the Federal standard deduction https://revisor.mo.gov/main/OneSection.aspx?section=143.131
        mo_standard_deduction = tax_unit("standard_deduction", period)
        mo_itemized_deductions = tax_unit("mo_itemized_deductions", period)
        mo_itemized_or_standard = where(
            tax_unit_itemizes, mo_itemized_deductions, mo_standard_deduction
        )
        # NB: The federal income tax deduction applies regardless of itemization.
        # Note: There would also be a personal and/or dependent exemptions as part of
        # this formula, but they are legally based on eligibility for the federal
        # versions of those exemptions, both of which are suspended through 2025 federally.
        return max_(
            mo_agi - mo_itemized_or_standard - mo_federal_income_tax_deduction,
            0,
        )
